# -*- coding: utf-8 -*-
# Copyright 2022 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Turbinia API client command-line tool."""

import click
import logging
import sys

from turbinia_api_lib import exceptions
from turbinia_client.factory import factory

log = logging.getLogger(__name__)


@click.group('config')
def config_group():
  """Get Turbinia configuration."""


@click.group('evidence')
def evidence_group():
  """Get or upload Turbinia evidence."""


@click.group('status')
def status_group():
  """Get Turbinia request or task status."""


@click.group('result')
def result_group():
  """Get Turbinia request or task results."""


@click.group('jobs')
def jobs_group():
  """Get a list of enabled Turbinia jobs."""


@click.group('report')
def report_group():
  """Get reports for Tasks or Requests."""


@click.group('logs')
def logs_group():
  """Get Turbinia logs."""


@click.group('submit')
@click.pass_context
def setup_submit(ctx: click.Context):
  try:
    ctx.obj.evidence_mapping = ctx.obj.get_evidence_arguments()
    ctx.obj.request_options = ctx.obj.get_request_options()
    # Build all the commands based on responses from the API server.
    request_commands = factory.CommandFactory.create_dynamic_objects(
        evidence_mapping=ctx.obj.evidence_mapping,
        request_options=ctx.obj.request_options)
    for command in request_commands:
      submit_group.add_command(command)
  except exceptions.ApiException as exception:
    log.error(
        'Error while attempting to contact the API server during setup: %s',
        exception)
    sys.exit(-1)


@click.group('submit', chain=True, invoke_without_command=True)
@click.pass_context
def submit_group(ctx: click.Context):
  """Submit new requests to the Turbinia API server.
  
  Please run this command without any arguments to view a list
  of available evidence types.
  """
  ctx.invoke(setup_submit)
  click.echo(submit_group.get_help(ctx))



#!/usr/bin/env python
#
# Copyright 2017 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Methods for formatting text."""

from __future__ import annotations

from abc import ABC, abstractmethod
from click import echo as click_echo
from httpx import Response
from typing import Any
from collections import defaultdict

import logging
import json
import pandas

MEDIUM_PRIORITY = 50
HIGH_PRIORITY = 20
CRITICAL_PRIORITY = 10

log = logging.getLogger(__name__)

IMPORTANT_ATTRIBUTES = {
    'id', '_name', 'type', 'size', 'request_id', 'tasks', 'source_path',
    'local_path', 'creation_time', 'last_update'
}


def echo_json(data: Any) -> None:
  """Pretty print JSON data."""
  try:
    if isinstance(data, str):
      json_string: str = json.loads(data)
      json_string = json.dumps(json_string, indent=2)
    else:
      json_string: str = json.dumps(data, indent=2)
    click_echo(json_string)
  except json.JSONDecodeError as exception:
    raise RuntimeError('Unable to decode API response') from exception


def decode_api_response(data: Any) -> str:
  """Decodes ApiResponse data into a Python object"""
  if isinstance(data, str) or isinstance(data, Response):
    return data

  data_attribute = None
  response = ''
  try:
    if data_attribute := getattr(data, 'data'):
      response = data_attribute
    if not data_attribute:
      if data_attribute := getattr(data, 'raw_data'):
        response = json.loads(data_attribute)
    return response
  except json.JSONDecodeError as exception:
    raise RuntimeError('Unable to decode API response') from exception
  except AttributeError as exception:
    raise RuntimeError('Unable to decode API response') from exception


class MarkdownReportComponent(ABC):
  """Components for generating Turbinia request/task
      markdown reports.
  """

  def __init__(self):
    """Instantiates a MarkdownReportComponent object."""
    self._components: list(MarkdownReportComponent) = []
    self._parent: MarkdownReportComponent = None
    self._report: str = None

  @property
  def components(self):
    """Returns the components list."""
    return self._components

  @property
  def report(self):
    """Returns the markdown report text."""
    return self._report

  @report.setter
  def report(self, report):
    self._report = report

  @property
  def parent(self) -> MarkdownReportComponent:
    """Returns the parent object."""
    return self._parent

  @parent.setter
  def parent(self, parent: MarkdownReportComponent):
    self._parent = parent

  def bold(self, text):
    """Formats text as bold in Markdown format.

    Args:
        text(string): Text to format

    Return:
        string: Formatted text.
    """
    return f'**{text.strip():s}**'

  def heading(self, text: str, number: int) -> str:
    """Formats text as heading in Markdown format.

    Args:
        text(string): Text to format
        number(int): Heading number

    Return:
        string: Formatted text.
    """
    return f'{"#"*number} {text.strip():s}'

  def heading1(self, text):
    """Formats text as heading 1 in Markdown format.

    Args:
        text(string): Text to format

    Return:
        string: Formatted text.
    """
    return self.heading(text, 1)

  def heading2(self, text):
    """Formats text as heading 2 in Markdown format.

    Args:
        text(string): Text to format

    Return:
        string: Formatted text.
    """
    return self.heading(text, 2)

  def heading3(self, text):
    """Formats text as heading 3 in Markdown format.

    Args:
        text(string): Text to format

    Return:
        string: Formatted text.
    """
    return self.heading(text, 3)

  def heading4(self, text):
    """Formats text as heading 4 in Markdown format.

    Args:
        text(string): Text to format

    Return:
        string: Formatted text.
    """
    return self.heading(text, 4)

  def heading5(self, text):
    """Formats text as heading 5 in Markdown format.
     Args:
        text(string): Text to format
     Return:
        string: Formatted text.
    """
    return self.heading(text, 5)

  def bullet(self, text, level=1):
    """Formats text as a bullet in Markdown format.

      Args:
        text(string): Text to format
        level(int): Indentation level.
      Return:
        string: Formatted text.
    """
    return f"{'    ' * (level - 1):s}* {text.strip():s}"

  def code(self, text):
    """Formats text as code in Markdown format.

      Args:
          text(string): Text to format

     Return:
          string: Formatted text.
    """
    return f'`{text.strip():s}`'

  def add(self, component: MarkdownReportComponent) -> None:
    """Adds a MarkdownReportComponent object to the components list.

    This method should additionally set the parent object.
    """
    pass

  def add_components(self, components: list[MarkdownReportComponent]) -> None:
    """Adds multiple MarkdownReportComponent objects to the components list."""
    pass

  def remove(self, component: MarkdownReportComponent) -> None:
    """Removes a MarkdownReportComponent object from the components list.

    This method should set the component's object to None.
    """
    pass

  def list_to_markdown(self, original_list: list, level: int = 1) -> list:
    """Generates a Markdown based on a python list.

      Args:
        original_list(list): List to format
        level(int): Indentation level.

      Returns:
        report (list): List of Markdown lines.
    """
    if not original_list:
      return [self.bullet('[EMPTY LIST]', level)]
    report = []
    for item in original_list:
      if isinstance(item, dict):
        report.extend(self.dict_to_markdown(item, level + 1))
      elif isinstance(item, list):
        report.extend(self.list_to_markdown(item, level + 1))
      else:
        if level == 0:
          report.append(self.heading1(item))
        else:
          report.append(self.bullet(item, level))
    return report

  def dict_to_markdown(
      self, original_dict: dict, level: int = 1, excluded_keys: list = (),
      format_keys: bool = True) -> list:
    """Generates a Markdown based on a python dictionary.

      Args:
        original_dict(dict): Dict to format.
        level(int): Indentation level.
        excluded_keys (list): List of dict keys to be ignored.
        format_keys (bool): Capitalizes and removes underscores from dict keys
          if True

      Returns:
        report (list): List of Markdown lines.
    """
    if not original_dict:
      return [self.bullet('[EMPTY DICTIONARY]', level)]
    report: list[str] = []
    for key, value in original_dict.items():
      if key in excluded_keys:
        continue
      name = key.replace('_', ' ').title() if format_keys else key
      if isinstance(value, dict):
        if level == 0:
          report.append(self.heading1(f'{name}:'))
        else:
          report.append(self.bullet(f'{name}:', level))
        report.extend(self.dict_to_markdown(value, level + 1))
      elif isinstance(value, list):
        if level == 0:
          report.append(self.heading1(f'{name}:'))
        else:
          report.append(self.bullet(f'{name}:', level))
        report.extend(self.list_to_markdown(value, level + 1))
      else:
        if level == 0:
          report.append(self.heading1(f'{name}: {value}'))
        else:
          report.append(self.bullet(f'{name}: {value}', level))
    return report

  @abstractmethod
  def generate_markdown(self) -> str:
    pass


class TaskMarkdownReport(MarkdownReportComponent):
  """Turbinia Task markdown report."""

  def __init__(self, request_data: dict = None):
    """Initialize TaskMarkdownReport"""
    super().__init__()
    self._request_data: dict = request_data

  def generate_markdown(
      self, priority_filter=None, show_all=False, compact=False) -> str:
    """Generate a markdown report."""
    report: list[str] = []
    task: dict = self._request_data
    if not task:
      return ''

    priority = task.get('report_priority') if task.get(
        'report_priority') else MEDIUM_PRIORITY
    priority_filter = priority_filter if priority_filter else HIGH_PRIORITY

    if priority <= CRITICAL_PRIORITY:
      name = f'{task.get("name")} ({"CRITICAL PRIORITY"})'
    elif priority <= HIGH_PRIORITY:
      name = f'{task.get("name")} ({"HIGH PRIORITY"})'
    elif priority <= MEDIUM_PRIORITY:
      name = f'{task.get("name")} ({"MEDIUM PRIORITY"})'
    else:
      name = f'{task.get("name")} ({"LOW PRIORITY"})'

    try:
      # Only show Task details if the Task has more priority than the
      # priority_filter
      if priority > priority_filter:
        report.append(f'{self.heading3(name)}: {task.get("status")!s}')
      else:
        report.append(self.heading2(name))
        line = f"{self.bold('Evidence:'):s} {task.get('evidence_name')!s}"
        report.append(self.bullet(line))
        line = f"{self.bold('Status:'):s} {task.get('status')!s}"
        report.append(self.bullet(line))

        report.append(self.bullet(f"Task Id: {task.get('id')!s}"))
        report.append(
            self.bullet(f"Executed on worker {task.get('worker_name')!s}"))

        if task.get('report_data'):
          if not compact:
            report.append('')
          report.append(self.heading3('Task Reported Data'))
          report.extend(task.get('report_data').splitlines())
        if not compact:
          report.append('')

      if show_all and priority <= priority_filter:
        if not compact:
          report.append('')
        report.append(self.heading3('Saved Task Files:'))
        saved_paths = task.get('saved_paths')
        if saved_paths:
          for path in saved_paths:
            report.append(self.bullet(self.code(path)))
            if not compact:
              report.append('')
        else:
          report.append('No saved files')

      if priority <= priority_filter:
        report.append('')

    except TypeError as exception:
      log.warning(f'Error formatting the Markdown report: {exception!s}')

    self.report = '\n'.join(report)
    return self.report


class RequestMarkdownReport(MarkdownReportComponent):
  """Turbinia Request Markdown report."""

  def __init__(self, request_data: dict):
    """Initializes a RequestMarkdownReport object."""
    super().__init__()
    self._request_data: dict = request_data

    sorted_tasks = sorted(
        request_data.get('tasks'), key=lambda x:
        (x['report_priority'] if x['report_priority'] else 0, x['name']))

    self.add_components([TaskMarkdownReport(task) for task in sorted_tasks])

  def add(self, component: MarkdownReportComponent) -> None:
    if component:
      self.components.append(component)
      component.parent = self

  def remove(self, component: MarkdownReportComponent) -> None:
    self.components.remove(component)
    component.parent = None

  def add_components(self, components: list[MarkdownReportComponent]) -> None:
    if components:
      for component in components:
        self.components.append(component)
        component.parent = self

  def generate_markdown(self, priority_filter=None, show_all=False) -> str:
    """Generates a Markdown version of Requests results."""
    report: list[str] = []
    request_dict: dict = self._request_data
    if not request_dict:
      return ''

    try:
      report.append(
          self.heading2(f"Request ID: {request_dict.get('request_id')}"))
      report.append(
          self.bullet(
              f"Last Update: {request_dict.get('last_task_update_time')}"))
      report.append(self.bullet(f"Requester: {request_dict.get('requester')}"))
      report.append(self.bullet(f"Reason: {request_dict.get('reason')}"))
      report.append(self.bullet(f"Status: {request_dict.get('status')}"))
      report.append(
          self.bullet(f"Failed tasks: {request_dict.get('failed_tasks')}"))
      report.append(
          self.bullet(f"Running tasks: {request_dict.get('running_tasks')}"))
      report.append(
          self.bullet(
              f"Successful tasks: {request_dict.get('successful_tasks')}"))
      report.append(
          self.bullet(f"Task Count: {request_dict.get('task_count')}"))
      report.append(
          self.bullet(f"Queued tasks: {request_dict.get('queued_tasks')}"))
      report.append(
          self.bullet(f"Evidence Name: {request_dict.get('evidence_name')}"))
      report.append(
          self.bullet(f"Evidence ID: {request_dict.get('evidence_id')}"))
      report.append('')
    except TypeError as exception:
      log.warning(f'Error formatting the Markdown report: {exception!s}')

    task_counter = defaultdict(int)
    unique_tasks = []
    for task in self.components:
      markdown = task.generate_markdown(
          priority_filter=priority_filter, show_all=show_all, compact=True)
      task_counter[markdown] += 1
      if markdown not in unique_tasks:
        unique_tasks.append(markdown)

    # Generate task list with counts
    for task in unique_tasks:
      if task_counter[task] > 1:
        report.append(f'{task} ({task_counter[task]}x)')
      else:
        report.append(task)

    self.report = '\n'.join(report)
    return self.report


class SummaryMarkdownReport(MarkdownReportComponent):
  """A markdown report summary of all Turbinia Requests."""

  def __init__(self, requests_summary: list[dict]):
    """Initialize SummaryMarkdownReport."""
    super().__init__()
    self._requests_summary = requests_summary

  def generate_markdown(self) -> str:
    """Generate a Markdown version of Requests summary results."""
    report: list[str] = []
    requests_status_list = None
    if self._requests_summary:
      requests_status_list = self._requests_summary.get('requests_status')

    if not requests_status_list:
      return '## No requests found.'

    sorted_requests = sorted(
        requests_status_list, key=lambda x: x['last_task_update_time'])
    for request_dict in sorted_requests:
      request_report = RequestMarkdownReport(request_dict).generate_markdown()
      report.append(request_report)

    self.report = '\n'.join(report)
    return self.report


class WorkersMarkdownReport(MarkdownReportComponent):
  """A markdown report of all tasks for a specific worker."""

  def __init__(self, workers_status: dict, days: int):
    super().__init__()
    self._workers_status: dict = workers_status
    self._days: int = days

  def generate_markdown(self) -> str:
    """Generates a Markdown version of tasks per worker.

    Returns:
      markdown (str): Markdown version of tasks per worker.
    """
    report = []
    worker_status = self._workers_status.copy()
    scheduled_tasks = worker_status.pop('scheduled_tasks')
    report.append(
        self.heading1(
            f'Turbinia report for Worker activity within {self._days} days'))
    report.append(self.bullet(f'{len(worker_status.keys())} Worker(s) found.'))
    report.append(
        self.bullet(
            f'{scheduled_tasks} Task(s) unassigned or scheduled and pending '
            f'Worker assignment.'))
    for worker_node, task_types in worker_status.items():
      report.append('')
      report.append(self.heading2(f'Worker Node: {worker_node:s}'))
      for task_type, tasks in task_types.items():
        report.append(self.heading3(task_type.replace('_', ' ').title()))
        if not tasks:
          report.append(self.bullet('No Tasks found.'))
          report.append('')
          continue
        for task_id, task_attributes in tasks.items():
          report.append(
              self.bullet(f'{task_id} - {task_attributes["task_name"]}'))
          for attribute_name, attribute_value in task_attributes.items():
            if attribute_name != 'task_name':
              formatted_name = attribute_name.replace('_', ' ').title()
              report.append(
                  self.bullet(f'{formatted_name}: {attribute_value}', level=2))
        report.append('')

    return '\n'.join(report)


class StatsMarkdownReport(MarkdownReportComponent):
  """A markdown report of the task statistics."""

  def __init__(self, statistics: dict):
    super().__init__()
    self._statistics: dict = statistics
    self.table_dict = {
        'TASK': [],
        'COUNT': [],
        'MIN': [],
        'MEAN': [],
        'MAX': []
    }

  def stat_to_row(self, task: str, stat_dict: dict):
    """Generates a row of the statistics table.

    Args:
      task (str): Name of the current task.
      stat_dict (dict): Dictionary with information about current row.
    """
    self.table_dict['TASK'].append(f'{task}')
    self.table_dict['COUNT'].append(stat_dict.get('count', ''))
    self.table_dict['MIN'].append(stat_dict.get('min', ''))
    self.table_dict['MEAN'].append(stat_dict.get('mean', ''))
    self.table_dict['MAX'].append(stat_dict.get('max', ''))

  def generate_data_frame(self) -> pandas.DataFrame:
    """Generates a pandas DataFrame of the statistics table

    Args:
      markdown (bool): Bool defining if the tasks should be in markdown format.

    Returns:
      data_frame (DataFrame): Statistics table in pandas DataFrame format.
    """
    for stat_group, stat_dict in self._statistics.items():
      stat_group = stat_group.replace('_', ' ').title()
      if stat_group in ('All Tasks', 'Successful Tasks', 'Failed Tasks',
                        'Requests'):
        first_column = stat_group
        self.stat_to_row(first_column, stat_dict)
        continue
      for description, inner_dict in stat_dict.items():
        first_column = f'{stat_group.split(" ")[-1]} {description}'
        self.stat_to_row(first_column, inner_dict)
    return pandas.DataFrame(self.table_dict)

  def generate_markdown(self) -> str:
    """Generates a Markdown version of task statistics.

    Returns:
      markdown(str): Markdown version of task statistics.
    """
    report = [self.heading1('Execution time statistics for Turbinia:')]
    data_frame = self.generate_data_frame()
    table = data_frame.to_markdown(index=False)
    report.append(table)
    return '\n'.join(report)

  def generate_csv(self) -> str:
    """Generates a CSV version of task statistics.

    Returns:
      csv(str): CSV version of task statistics.
    """
    return self.generate_data_frame().to_csv(index=False)


class EvidenceMarkdownReport(MarkdownReportComponent):
  """Turbinia Evidence Markdown report."""

  def __init__(self, evidence_data: dict):
    """Initializes an EvidenceMarkdownReport object."""
    super().__init__()
    self._evidence_data: dict = evidence_data

  def generate_markdown(self, level=1, show_all=False) -> str:
    """Generates a Markdown version of Evidence.

      Args:
        level (int): Indentation level.
        show_all (bool): Shows all evidence attributes if True.

      Returns:
        report (str): Markdown report.
    """
    report: list[str] = []
    evidence_dict: dict = self._evidence_data
    if not evidence_dict:
      return ''
    try:
      report.append(
          self.heading(
              f"Evidence ID: {evidence_dict.get('id', 'null')}", level))
      report.append(
          self.bullet(
              f"Evidence Name: {evidence_dict.get('_name', 'null')}", level))
      report.append(
          self.bullet(
              f"Evidence Type: {evidence_dict.get('type', 'null')}", level))
      report.append(
          self.bullet(
              f"Evidence Size: {evidence_dict.get('size', 'null')}", level))
      report.append(
          self.bullet(
              f"Request ID: {evidence_dict.get('request_id', 'null')}", level))
      report.append(self.bullet('Tasks:', level))
      report.extend(
          self.list_to_markdown(evidence_dict.get('tasks'), level + 1))
      report.append(
          self.bullet(
              f"Source Path: {evidence_dict.get('source_path', 'null')}",
              level))
      report.append(
          self.bullet(
              f"Local Path: {evidence_dict.get('local_path', 'null')}", level))
      report.append(
          self.bullet(
              f"Creation Time: {evidence_dict.get('creation_time', 'null')}",
              level))
      report.append(
          self.bullet(
              f"Last Update: {evidence_dict.get('last_update', 'null')}",
              level))

      if show_all:
        report.extend(
            self.dict_to_markdown(
                evidence_dict, level=level, excluded_keys=IMPORTANT_ATTRIBUTES))

      report.append('')

    except TypeError as exception:
      log.warning(f'Error formatting the Markdown report: {exception!s}')

    self.report = '\n'.join(report)
    return self.report


class EvidenceSummaryMarkdownReport(EvidenceMarkdownReport):
  """Turbinia Evidence Markdown report."""

  def __init__(self, summary: dict | list | int):
    """Initializes an EvidenceSummaryMarkdownReport object."""
    super().__init__({})
    self._summary = summary

  def generate_content_markdown(self, summary, level=1):
    """Generates the content Markdown summary.

      Args:
        summary (bool): Evidence summary.
        level (int): Indentation level.

      Returns:
        report (list): List with Markdown lines.
    """
    report = []
    for item in summary:
      self._evidence_data = item
      report.append(self.generate_markdown(level=(level + 1)))
    return report

  def generate_summary_markdown(self, output: str = 'keys') -> str:
    """Generates the evidence summary Markdown.

      Args:
        output (str):Type of output (keys | content | count).

      Returns:
        report (str): Markdown report.
    """
    if output == 'content':
      if isinstance(self._summary, dict):
        report = []
        for attribute, summary in self._summary.items():
          report.append(self.bullet(f'{attribute}:'))
          report.extend(self.generate_content_markdown(summary, level=2))
        return '\n'.join(report)
      return '\n'.join(self.generate_content_markdown(self._summary))
    elif isinstance(self._summary, list):
      return '\n'.join(self.list_to_markdown(self._summary, level=0))
    elif isinstance(self._summary, dict):
      return '\n'.join(
          self.dict_to_markdown(self._summary, level=0, format_keys=False))
    elif isinstance(self._summary, int):
      return self.heading1(f'{self._summary} evidences found')



# -*- coding: utf-8 -*-
# Copyright 2023 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Turbinia API server unit tests."""

import unittest

from turbinia.api.cli.turbinia_client.helpers import formatter

from textwrap import dedent


class TestTurbiniaAPIFormatter(unittest.TestCase):
  """ Test Turbinia API formatter."""

  WORKERS_API_RESPONSE = {
      'scheduled_tasks': 0,
      '58be0171c8f0': {
          'run_status': {
              '950ef885ec32490ea6ee080b1f646a53': {
                  'task_name': 'GrepTask',
                  'last_update': '2023-08-31T01:54:08.104314Z',
                  'status': 'Task GrepTask is running on 58be0171c8f0',
                  'run_time': '0:00:03.418060'
              },
              '8c7d486bdb7444cda564d308e0d13476': {
                  'task_name':
                      'FinalizeRequestTask',
                  'last_update':
                      '2023-08-31T01:54:18.108941Z',
                  'status':
                      'Task FinalizeRequestTask is running on 58be0171c8f0',
                  'run_time':
                      '0:00:01.807367'
              }
          },
          'queued_status': {},
          'other_status': {
              '89ab31b41d8b4c30b007b09139d23695': {
                  'task_name':
                      'PsortTask',
                  'last_update':
                      '2023-08-31T01:54:08.085727Z',
                  'status':
                      'Completed successfully in 0:00:01.940446 on 58be0171c8f0',
                  'run_time':
                      '0:00:01.940446'
              },
              '30ea24b95a6848faba91893605a9b143': {
                  'task_name': 'WordpressCredsAnalysisTask',
                  'last_update': '2023-08-31T01:54:08.072453Z',
                  'status': 'No weak passwords found',
                  'run_time': '0:00:04.259982'
              }
          }
      }
  }

  EXPECTED_WORKERS_OUTPUT = dedent(
      """\
      # Turbinia report for Worker activity within 7 days
      * 1 Worker(s) found.
      * 0 Task(s) unassigned or scheduled and pending Worker assignment.

      ## Worker Node: 58be0171c8f0
      ### Run Status
      * 950ef885ec32490ea6ee080b1f646a53 - GrepTask
          * Last Update: 2023-08-31T01:54:08.104314Z
          * Status: Task GrepTask is running on 58be0171c8f0
          * Run Time: 0:00:03.418060
      * 8c7d486bdb7444cda564d308e0d13476 - FinalizeRequestTask
          * Last Update: 2023-08-31T01:54:18.108941Z
          * Status: Task FinalizeRequestTask is running on 58be0171c8f0
          * Run Time: 0:00:01.807367

      ### Queued Status
      * No Tasks found.

      ### Other Status
      * 89ab31b41d8b4c30b007b09139d23695 - PsortTask
          * Last Update: 2023-08-31T01:54:08.085727Z
          * Status: Completed successfully in 0:00:01.940446 on 58be0171c8f0
          * Run Time: 0:00:01.940446
      * 30ea24b95a6848faba91893605a9b143 - WordpressCredsAnalysisTask
          * Last Update: 2023-08-31T01:54:08.072453Z
          * Status: No weak passwords found
          * Run Time: 0:00:04.259982
      """)

  STATISTICS_API_RESPONSE = {
      'all_tasks': {
          'count': 18,
          'min': '0:00:00',
          'mean': '0:00:04',
          'max': '0:00:13'
      },
      'successful_tasks': {
          'count': 18,
          'min': '0:00:00',
          'mean': '0:00:04',
          'max': '0:00:13'
      },
      'failed_tasks': {
          'count': 0,
          'min': 'None',
          'mean': 'None',
          'max': 'None'
      },
      'requests': {
          'count': 2,
          'min': '0:01:06',
          'mean': '0:01:07',
          'max': '0:01:07'
      },
      'tasks_per_type': {
          'FinalizeRequestTask': {
              'count': 2,
              'min': '0:00:00',
              'mean': '0:00:00',
              'max': '0:00:00'
          },
          'GrepTask': {
              'count': 2,
              'min': '0:00:00',
              'mean': '0:00:00',
              'max': '0:00:00'
          },
          'LinuxAccountAnalysisTask': {
              'count': 2,
              'min': '0:00:06',
              'mean': '0:00:07',
              'max': '0:00:07'
          },
          'PlasoHasherTask': {
              'count': 2,
              'min': '0:00:11',
              'mean': '0:00:12',
              'max': '0:00:12'
          },
          'PlasoParserTask': {
              'count': 2,
              'min': '0:00:11',
              'mean': '0:00:13',
              'max': '0:00:13'
          },
          'PostgresAccountAnalysisTask': {
              'count': 2,
              'min': '0:00:03',
              'mean': '0:00:04',
              'max': '0:00:04'
          },
          'PsortTask': {
              'count': 2,
              'min': '0:00:01',
              'mean': '0:00:01',
              'max': '0:00:01'
          },
          'WindowsAccountAnalysisTask': {
              'count': 2,
              'min': '0:00:04',
              'mean': '0:00:05',
              'max': '0:00:05'
          },
          'WordpressCredsAnalysisTask': {
              'count': 2,
              'min': '0:00:03',
              'mean': '0:00:04',
              'max': '0:00:04'
          }
      },
      'tasks_per_worker': {
          '58be0171c8f0': {
              'count': 18,
              'min': '0:00:00',
              'mean': '0:00:04',
              'max': '0:00:13'
          }
      },
      'tasks_per_user': {
          'root': {
              'count': 18,
              'min': '0:00:00',
              'mean': '0:00:04',
              'max': '0:00:13'
          }
      }
  }

  # pylint: disable=line-too-long
  EXPECTED_STATISTICS_RESPONSE = dedent(
      """
      # Execution time statistics for Turbinia:
      | TASK                             |   COUNT | MIN     | MEAN    | MAX     |
      |:---------------------------------|--------:|:--------|:--------|:--------|
      | All Tasks                        |      18 | 0:00:00 | 0:00:04 | 0:00:13 |
      | Successful Tasks                 |      18 | 0:00:00 | 0:00:04 | 0:00:13 |
      | Failed Tasks                     |       0 | None    | None    | None    |
      | Requests                         |       2 | 0:01:06 | 0:01:07 | 0:01:07 |
      | Type FinalizeRequestTask         |       2 | 0:00:00 | 0:00:00 | 0:00:00 |
      | Type GrepTask                    |       2 | 0:00:00 | 0:00:00 | 0:00:00 |
      | Type LinuxAccountAnalysisTask    |       2 | 0:00:06 | 0:00:07 | 0:00:07 |
      | Type PlasoHasherTask             |       2 | 0:00:11 | 0:00:12 | 0:00:12 |
      | Type PlasoParserTask             |       2 | 0:00:11 | 0:00:13 | 0:00:13 |
      | Type PostgresAccountAnalysisTask |       2 | 0:00:03 | 0:00:04 | 0:00:04 |
      | Type PsortTask                   |       2 | 0:00:01 | 0:00:01 | 0:00:01 |
      | Type WindowsAccountAnalysisTask  |       2 | 0:00:04 | 0:00:05 | 0:00:05 |
      | Type WordpressCredsAnalysisTask  |       2 | 0:00:03 | 0:00:04 | 0:00:04 |
      | Worker 58be0171c8f0              |      18 | 0:00:00 | 0:00:04 | 0:00:13 |
      | User root                        |      18 | 0:00:00 | 0:00:04 | 0:00:13 |
      """)

  def testWorkersMarkdownReport(self):
    """Test formatting workers Markdown report."""
    result = formatter.WorkersMarkdownReport(self.WORKERS_API_RESPONSE,
                                             7).generate_markdown()

    self.assertEqual(result, self.EXPECTED_WORKERS_OUTPUT)

  def testStatisticsMarkdownReport(self):
    """Test formatting statistics Markdown report."""
    result = formatter.StatsMarkdownReport(
        self.STATISTICS_API_RESPONSE).generate_markdown()

    self.assertEqual(result, self.EXPECTED_STATISTICS_RESPONSE.strip())



# -*- coding: utf-8 -*-
# Copyright 2022 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Turbinia API client command-line tool."""

import os
import logging
import click
import tarfile

from importlib.metadata import version as importlib_version

from turbinia_api_lib import exceptions
from turbinia_api_lib import api_client
from turbinia_api_lib import models
from turbinia_api_lib.api import turbinia_requests_api
from turbinia_api_lib.api import turbinia_tasks_api
from turbinia_api_lib.api import turbinia_configuration_api
from turbinia_api_lib.api import turbinia_jobs_api
from turbinia_api_lib.api import turbinia_request_results_api
from turbinia_api_lib.api import turbinia_evidence_api
from turbinia_api_lib.api import turbinia_logs_api
from turbinia_client.core import groups
from turbinia_client.helpers import formatter

log = logging.getLogger(__name__)


@groups.config_group.command('list')
@click.pass_context
def get_config(ctx: click.Context) -> None:
  """Gets Turbinia server configuration."""
  client: api_client.ApiClient = ctx.obj.api_client
  api_instance = turbinia_configuration_api.TurbiniaConfigurationApi(client)
  try:
    api_response = api_instance.read_config_with_http_info()
    decoded_response = formatter.decode_api_response(api_response)
    formatter.echo_json(decoded_response)
  except exceptions.ApiException as exception:
    log.error(
        f'Received status code {exception.status} '
        f'when calling read_config_with_http_info: {exception.body}')


@groups.config_group.command('version')
@click.pass_context
def get_api_server_version(ctx: click.Context) -> None:
  """Gets Turbinia API server version."""
  client: api_client.ApiClient = ctx.obj.api_client
  api_instance = turbinia_configuration_api.TurbiniaConfigurationApi(client)
  try:
    api_response = api_instance.get_version()
    decoded_response = formatter.decode_api_response(api_response)
    formatter.echo_json(decoded_response)
  except exceptions.ApiException as exception:
    log.error(
        f'Received status code {exception.status} '
        f'when calling get_version: {exception.body}')


@groups.config_group.command('download')
@click.pass_context
def get_config_download(ctx: click.Context) -> None:
  """Downloads current Turbinia config from the API server."""
  client: api_client.ApiClient = ctx.obj.api_client
  api_instance = turbinia_configuration_api.TurbiniaConfigurationApi(client)
  try:
    api_response = api_instance.download_config()
    click.echo(api_response)
  except exceptions.ApiException as exception:
    log.error(
        f'Received status code {exception.status} '
        f'when calling download_config: {exception.body}')


@groups.result_group.command('request')
@click.pass_context
@click.argument('request_id')
def get_request_result(ctx: click.Context, request_id: str) -> None:
  """Gets Turbinia request results / output files."""
  client: api_client.ApiClient = ctx.obj.api_client
  api_instance = turbinia_request_results_api.TurbiniaRequestResultsApi(client)
  filename = f'{request_id}.tgz'
  click.echo(f'Downloading output for request {request_id} to: {filename}')
  try:
    api_response = api_instance.get_request_output_with_http_info(
        request_id, _preload_content=False, _request_timeout=(30, 900))
    # Read the response and save into a local file.
    with open(filename, 'wb') as file:
      file.write(api_response.raw_data)
  except exceptions.ApiException as exception:
    log.error(
        f'Received status code {exception.status} '
        f'when calling get_request_output_with_http_info: {exception.body}')
  except OSError as exception:
    log.error(f'Unable to save file: {exception}')
  except (ValueError, tarfile.ReadError, tarfile.CompressionError) as exception:
    log.error(f'Error reading saved results file {filename}: {exception}')


@groups.result_group.command('task')
@click.pass_context
@click.argument('task_id')
def get_task_result(ctx: click.Context, task_id: str) -> None:
  """Gets Turbinia task results / output files."""
  client: api_client.ApiClient = ctx.obj.api_client
  api_instance = turbinia_request_results_api.TurbiniaRequestResultsApi(client)
  filename = f'{task_id}.tgz'
  click.echo(f'Downloading output for task {task_id} to: {filename}')
  try:
    api_response = api_instance.get_task_output_with_http_info(
        task_id, _preload_content=False, _request_timeout=(30, 900))
    # Read the response and save into a local file.
    with open(filename, 'wb') as file:
      file.write(api_response.raw_data)
  except exceptions.ApiException as exception:
    log.error(
        f'Received status code {exception.status} '
        f'when calling get_task_output_with_http_info: {exception.body}')
  except OSError as exception:
    log.error(f'Unable to save file: {exception}')
  except (ValueError, tarfile.ReadError, tarfile.CompressionError) as exception:
    log.error(f'Error reading saved results file {filename}: {exception}')


@groups.result_group.command('plasofile')
@click.pass_context
@click.argument('task_id')
def get_plaso_file(ctx: click.Context, task_id: str) -> None:
  """Gets Turbinia task results / output files."""
  client: api_client.ApiClient = ctx.obj.api_client
  api_instance = turbinia_request_results_api.TurbiniaRequestResultsApi(client)
  filename = f'{task_id}.plaso'
  click.echo(f'Downloading output for task {task_id} to: {filename}')
  try:
    api_response = api_instance.get_plaso_file_with_http_info(
        task_id, _preload_content=False, _request_timeout=(30, 900))
    # Read the response and save into a local file.
    with open(filename, 'wb') as file:
      file.write(api_response.raw_data)
  except exceptions.ApiException as exception:
    log.error(
        f'Received status code {exception.status} '
        f'when calling get_plaso_file_with_http_info: {exception.body}')
  except OSError as exception:
    log.error(f'Unable to save file: {exception}')
  except (ValueError, tarfile.ReadError, tarfile.CompressionError) as exception:
    log.error(f'Error reading saved results file {filename}: {exception}')


@groups.jobs_group.command('list')
@click.pass_context
def get_jobs(ctx: click.Context) -> None:
  """Gets Turbinia jobs list."""
  client: api_client.ApiClient = ctx.obj.api_client
  api_instance = turbinia_jobs_api.TurbiniaJobsApi(client)
  try:
    api_response = api_instance.read_jobs_with_http_info()
    decoded_response = formatter.decode_api_response(api_response)
    formatter.echo_json(decoded_response)
  except exceptions.ApiException as exception:
    log.error(
        f'Received status code {exception.status} '
        f'when calling read_jobs_with_http_info: {exception.body}')


@groups.status_group.command('request')
@click.pass_context
@click.argument('request_id')
@click.option(
    '--priority_filter', '-p', help='This sets what report sections are '
    'shown in full detail in report output.  Any tasks that have set a '
    'report_priority value equal to or lower than this setting will be '
    'shown in full detail, and tasks with a higher value will only have '
    'a summary shown.  The default is 20 which corresponds to "HIGH_PRIORITY"'
    'To see all tasks report output in full detail, set --priority_filter=100 '
    'or to see CRITICAL only set --priority_filter=10', show_default=True,
    default=20, type=int, required=False)
@click.option(
    '--show_all', '-a', help='Shows all fields including saved output paths.',
    is_flag=True, required=False)
@click.option(
    '--json_dump', '-j', help='Generates JSON output.', is_flag=True,
    required=False)
def get_request(
    ctx: click.Context, request_id: str, priority_filter: int, show_all: bool,
    json_dump: bool) -> None:
  """Gets Turbinia request status."""
  client: api_client.ApiClient = ctx.obj.api_client
  api_instance = turbinia_requests_api.TurbiniaRequestsApi(client)
  if request_id == 'summary':
    click.echo(
        'Oops! "summary" is not a valid request identifier. '
        'Did you mean to run "turbinia-client status summary" instead?')
    return
  try:
    api_response = api_instance.get_request_status_with_http_info(request_id)
    decoded_response = formatter.decode_api_response(api_response)
    if json_dump:
      formatter.echo_json(decoded_response)
    else:
      report = formatter.RequestMarkdownReport(
          decoded_response).generate_markdown(
              priority_filter=priority_filter, show_all=show_all)
      click.echo(report)
  except exceptions.ApiException as exception:
    log.error(
        f'Received status code {exception.status} '
        f'when calling get_request_status_with_http_info: {exception.body}')


@groups.status_group.command('workers')
@click.pass_context
@click.option(
    '--days', '-d', help='Specifies status timeframe.', required=False)
@click.option(
    '--all_fields', '-a', help='Returns all fields.', is_flag=True,
    required=False)
@click.option(
    '--json_dump', '-j', help='Generates JSON output.', is_flag=True,
    required=False)
def get_workers(
    ctx: click.Context, days: int, all_fields: bool, json_dump: bool) -> None:
  """Shows Workers status information."""
  days = int(days) if days else 7
  client: api_client.ApiClient = ctx.obj.api_client
  api_instance = turbinia_tasks_api.TurbiniaTasksApi(client)
  try:
    api_response = api_instance.get_workers_status_with_http_info(
        days, all_fields)
    decoded_response = formatter.decode_api_response(api_response)

    if json_dump:
      formatter.echo_json(decoded_response)
    else:
      report = formatter.WorkersMarkdownReport(decoded_response,
                                               days).generate_markdown()
      click.echo(report)
  except exceptions.ApiException as exception:
    log.error(
        f'Received status code {exception.status} '
        f'when calling get_workers_status_with_http_info: {exception.body}')


@groups.status_group.command('statistics')
@click.pass_context
@click.option(
    '--days', '-d', help='Specifies statistics timeframe.', required=False)
@click.option(
    '--task_id', '-t', help='Gets statistics for a specific task.',
    required=False)
@click.option(
    '--request_id', '-r', help='Gets statistics for a specific request.',
    required=False)
@click.option(
    '--user', '-u', help='Gets statistics for a specific user.', required=False)
@click.option(
    '--csv', '-c', help='Outputs statistics as CSV.', is_flag=True,
    required=False)
@click.option(
    '--json_dump', '-j', help='Generates JSON output.', is_flag=True,
    required=False)
def get_statistics(
    ctx: click.Context, days: int, task_id: str, request_id: str, user: str,
    csv: bool, json_dump: bool) -> None:
  """Shows statistics about tasks."""
  days = int(days) if days else 7
  client: api_client.ApiClient = ctx.obj.api_client
  api_instance = turbinia_tasks_api.TurbiniaTasksApi(client)
  try:
    api_response = api_instance.get_task_statistics_with_http_info(
        days=days, task_id=task_id, request_id=request_id, user=user)
    decoded_response: models.CompleteTurbiniaStats = (
        formatter.decode_api_response(api_response))
    if json_dump:
      formatter.echo_json(decoded_response)
    else:
      stat_formatter = formatter.StatsMarkdownReport(decoded_response.dict())
      if csv:
        report = stat_formatter.generate_csv()
      else:
        report = stat_formatter.generate_markdown()
      click.echo(report)
  except exceptions.ApiException as exception:
    log.error(
        f'Received status code {exception.status} '
        f'when calling get_task_statistics_with_http_info: {exception.body}')


@groups.status_group.command('summary')
@click.pass_context
@click.option(
    '--json_dump', '-j', help='Generates JSON output.', is_flag=True,
    required=False)
def get_requests_summary(ctx: click.Context, json_dump: bool) -> None:
  """Gets a summary of all Turbinia requests."""
  client: api_client.ApiClient = ctx.obj.api_client
  api_instance = turbinia_requests_api.TurbiniaRequestsApi(client)
  try:
    api_response = api_instance.get_requests_summary_with_http_info()
    decoded_response = formatter.decode_api_response(api_response)

    if json_dump:
      formatter.echo_json(decoded_response)
    else:
      report = formatter.SummaryMarkdownReport(
          decoded_response).generate_markdown()
      click.echo(report)
  except exceptions.ApiException as exception:
    log.error(
        f'Received status code {exception.status} '
        f'when calling get_requests_summary_with_http_info: {exception.body}')


@groups.status_group.command('task')
@click.pass_context
@click.argument('task_id')
@click.option(
    '--show_all', '-a', help='Shows all field regardless of priority.',
    is_flag=True, required=False)
@click.option(
    '--json_dump', '-j', help='Generates JSON output.', is_flag=True,
    required=False)
def get_task(
    ctx: click.Context, task_id: str, show_all: bool, json_dump: bool) -> None:
  """Gets Turbinia task status."""
  client: api_client.ApiClient = ctx.obj.api_client
  api_instance = turbinia_tasks_api.TurbiniaTasksApi(client)
  try:
    api_response = api_instance.get_task_status_with_http_info(task_id)
    decoded_response = formatter.decode_api_response(api_response)
    if json_dump:
      formatter.echo_json(decoded_response)
    else:
      report = formatter.TaskMarkdownReport(decoded_response).generate_markdown(
          show_all=show_all, priority_filter=100)
      click.echo(report)
  except exceptions.ApiException as exception:
    log.error(
        f'Received status code {exception.status} '
        f'when calling get_task_status_with_http_info: {exception.body}')


@groups.evidence_group.command('summary')
@click.pass_context
@click.option(
    '--group', '-g', help='Attribute by which output will be grouped.',
    default=None, required=False)
@click.option(
    '--output', '-o', help='Type of output (keys | content | count).',
    default='keys', required=False)
@click.option(
    '--json_dump', '-j', help='Generates JSON output.', is_flag=True,
    required=False)
def get_evidence_summary(
    ctx: click.Context, group: str, output: str, json_dump: bool) -> None:
  """Gets Turbinia evidence summary."""
  client: api_client.ApiClient = ctx.obj.api_client
  api_instance = turbinia_evidence_api.TurbiniaEvidenceApi(client)
  try:
    api_response = api_instance.get_evidence_summary_with_http_info(
        group, output)
    decoded_response = formatter.decode_api_response(api_response)

    if json_dump:
      formatter.echo_json(decoded_response)
    else:
      report = formatter.EvidenceSummaryMarkdownReport(
          decoded_response).generate_summary_markdown(output)
      click.echo(report)
  except exceptions.ApiException as exception:
    log.error(
        f'Received status code {exception.status} '
        f'when calling get_evidence_summary_with_http_info: {exception.body}')


@groups.evidence_group.command('query')
@click.pass_context
@click.argument('attribute_name')
@click.argument('attribute_value')
@click.option(
    '--output', '-o', help='Type of output (keys | content | count).',
    default='keys', is_flag=False, required=False)
@click.option(
    '--json_dump', '-j', help='Generates JSON output.', is_flag=True,
    required=False)
def query_evidence(
    ctx: click.Context, attribute_name: str, attribute_value: str, output: str,
    json_dump: bool) -> None:
  """Queries Turbinia evidence."""
  client: api_client.ApiClient = ctx.obj.api_client
  api_instance = turbinia_evidence_api.TurbiniaEvidenceApi(client)
  try:
    api_response = api_instance.query_evidence_with_http_info(
        attribute_value, attribute_name, output)
    decoded_response = formatter.decode_api_response(api_response)

    if json_dump:
      formatter.echo_json(decoded_response)
    else:
      report = formatter.EvidenceSummaryMarkdownReport(
          decoded_response).generate_summary_markdown(output)
      click.echo(report)
  except exceptions.ApiException as exception:
    log.error(
        f'Received status code {exception.status} '
        f'when calling query_evidence_with_http_info: {exception.body}')


@groups.evidence_group.command('get')
@click.pass_context
@click.argument('evidence_id')
@click.option(
    '--show_all', '-a', help='Shows all evidence attributes.', is_flag=True,
    required=False)
@click.option(
    '--json_dump', '-j', help='Generates JSON output.', is_flag=True,
    required=False)
def get_evidence(
    ctx: click.Context, evidence_id: str, show_all: bool,
    json_dump: bool) -> None:
  """Get Turbinia evidence."""
  client: api_client.ApiClient = ctx.obj.api_client
  api_instance = turbinia_evidence_api.TurbiniaEvidenceApi(client)
  try:
    api_response = api_instance.get_evidence_by_id_with_http_info(evidence_id)
    decoded_response = formatter.decode_api_response(api_response)

    if json_dump:
      formatter.echo_json(decoded_response)
    else:
      report = formatter.EvidenceMarkdownReport(
          decoded_response).generate_markdown(1, show_all=show_all)
      click.echo(report)
  except exceptions.ApiException as exception:
    log.error(
        f'Received status code {exception.status} '
        f'when calling get_evidence_by_id_with_http_info: {exception.body}')


@groups.evidence_group.command('upload')
@click.pass_context
@click.argument('ticket_id')
@click.option(
    '--path', '-p', help='Path of file or directory to be uploaded.',
    required=True, multiple=True)
@click.option(
    '--calculate_hash', '-c', help='Calculates file hash.', is_flag=True,
    required=False)
@click.option(
    '--json_dump', '-j', help='Generates JSON output.', is_flag=True,
    required=False)
def upload_evidence(
    ctx: click.Context, ticket_id: str, path: list, calculate_hash: bool,
    json_dump: bool) -> None:
  """Uploads evidence to Turbinia server."""
  client: api_client.ApiClient = ctx.obj.api_client
  api_instance_config = turbinia_configuration_api.TurbiniaConfigurationApi(
      client)
  max_upload_size = api_instance_config.read_config()['API_MAX_UPLOAD_SIZE']
  api_instance = turbinia_evidence_api.TurbiniaEvidenceApi(client)
  files = []
  for current_path in path:
    if os.path.isfile(current_path):
      files.append(current_path)
      continue
    for file_name in os.listdir(current_path):
      file_path = os.path.join(current_path, file_name)
      if os.path.isfile(file_path):
        files.append(file_path)
  report = {}
  for file_path in files:
    try:
      size = os.path.getsize(file_path)
      if size > max_upload_size:
        error_message = (
            f'Unable to upload {size / (1024 ** 3)} GB file',
            f'{file_path} greater than {max_upload_size / (1024 ** 3)} GB')
        log.error(error_message)
        continue
      abs_path = os.path.abspath(file_path)
    except OSError:
      log.error(f'Unable to read file in {file_path}')
      continue
    try:
      api_response = api_instance.upload_evidence_with_http_info(
          [file_path], ticket_id, calculate_hash)
      report[abs_path] = formatter.decode_api_response(api_response)
    except exceptions.ApiException as exception:
      error_message = (
          f'Received status code {exception.status} '
          f'when calling upload_evidence_with_http_info: {exception}')
      log.error(error_message)
      report[abs_path] = error_message
  if json_dump:
    formatter.echo_json(report)
  else:
    report = '\n'.join(
        formatter.EvidenceMarkdownReport({}).dict_to_markdown(
            report, 0, format_keys=False))
    click.echo(report)


@click.command('version')
def version():
  """Returns the turbinia-client package distribution version."""
  cli_version = importlib_version('turbinia-client')
  click.echo(f'turbinia-client version {cli_version}')


@groups.report_group.command('request')
@click.pass_context
@click.argument('request_id')
def get_request_report(ctx: click.Context, request_id: str) -> None:
  """Gets Turbinia request Markdown report.
  
  This command will not do any filtering since this is provided in
  the `status request` command via --priority_filter

  """
  client: api_client.ApiClient = ctx.obj.api_client
  api_instance = turbinia_requests_api.TurbiniaRequestsApi(client)
  try:
    api_response = api_instance.get_request_report(request_id)
    click.echo(api_response)
  except exceptions.ApiException as exception:
    log.error(
        f'Received status code {exception.status} '
        f'when calling get_request_report: {exception.body}')


@groups.report_group.command('task')
@click.pass_context
@click.argument('task_id')
def get_task_report(ctx: click.Context, task_id: str) -> None:
  """Gets Turbinia task Markdown report."""
  client: api_client.ApiClient = ctx.obj.api_client
  api_instance = turbinia_tasks_api.TurbiniaTasksApi(client)
  try:
    api_response = api_instance.get_task_report(task_id)
    click.echo(api_response)
  except exceptions.ApiException as exception:
    log.error(
        f'Received status code {exception.status} '
        f'when calling get_request_report: {exception.body}')


@groups.logs_group.command('system')
@click.pass_context
@click.argument('hostname')
@click.option(
    '--num_lines', '-n', help='Maximum number of log lines to retrieve.',
    required=False, is_flag=False, default=500)
def get_logs(ctx: click.Context, hostname: str, num_lines: int) -> None:
  """Gets Turbinia system logs."""
  client: api_client.ApiClient = ctx.obj.api_client
  api_instance = turbinia_logs_api.TurbiniaLogsApi(client)
  try:
    api_response = api_instance.get_turbinia_logs(hostname, num_lines)
    click.echo(api_response)
  except exceptions.ApiException as exception:
    log.error(
        f'Received status code {exception.status} '
        f'when calling get_turbinia_logs: {exception.body}')


@groups.logs_group.command('api')
@click.pass_context
@click.option(
    '--num_lines', '-n', help='Maximum number of log lines to retrieve.',
    required=False, is_flag=False, default=500)
def get_api_server_logs(ctx: click.Context, num_lines: int) -> None:
  """Gets Turbinia system logs."""
  client: api_client.ApiClient = ctx.obj.api_client
  api_instance = turbinia_logs_api.TurbiniaLogsApi(client)
  try:
    api_response = api_instance.get_api_server_logs(num_lines)
    click.echo(api_response)
  except exceptions.ApiException as exception:
    log.error(
        f'Received status code {exception.status} '
        f'when calling get_turbinia_logs: {exception.body}')



# -*- coding: utf-8 -*-
# Copyright 2022 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Turbinia API client command-line tool."""

import os
import logging
import click
import base64

from typing import Tuple, Sequence
from turbinia_api_lib.api import turbinia_configuration_api
from turbinia_api_lib.api import turbinia_requests_api

from turbinia_api_lib import exceptions
from turbinia_api_lib import api_client

log = logging.getLogger(__name__)


def generate_option_parameters(
    option_name: str) -> Tuple[Tuple[Sequence[str], str], dict]:
  """Builds click.Option or click.Command arguments based on a given parameter 
      name.
  """
  return ((['--' + option_name], option_name), {'required': False, 'type': str})


@click.pass_context
def create_request(ctx: click.Context, *args: int, **kwargs: int) -> None:
  """Creates and submits a new Turbinia request."""
  client: api_client.ApiClient = ctx.obj.api_client
  api_instance = turbinia_requests_api.TurbiniaRequestsApi(client)
  evidence_name = ctx.command.name

  # Normalize the evidence class name from lowercase to the original name.
  evidence_name = ctx.obj.normalize_evidence_name(evidence_name)
  # Build request and request_options objects to send to the API server.
  request_options = list(ctx.obj.request_options.keys())
  request = {'evidence': {'type': evidence_name}, 'request_options': {}}

  if 'googlecloud' in evidence_name:
    api_instance_config = turbinia_configuration_api.TurbiniaConfigurationApi(
        client)
    cloud_provider = api_instance_config.read_config()['CLOUD_PROVIDER']
    if cloud_provider != 'GCP':
      log.error(
          f'The evidence type {evidence_name} is Google Cloud only and '
          f'the configured provider for this Turbinia instance is '
          f'{cloud_provider}.')
      return

  for key, value in kwargs.items():
    # If the value is not empty, add it to the request.
    if kwargs.get(key):
      # Check if the key is for evidence or request_options
      if not key in request_options:
        request['evidence'][key] = value
      elif key in ('jobs_allowlist', 'jobs_denylist'):
        jobs_list = value.split(',')
        request['request_options'][key] = jobs_list
      else:
        request['request_options'][key] = value

  if all(key in request['request_options']
         for key in ('recipe_name', 'recipe_data')):
    log.error('You can only provide one of recipe_data or recipe_name')
    return

  recipe_name = request['request_options'].get('recipe_name')
  if recipe_name:
    if not recipe_name.endswith('.yaml'):
      recipe_name = f'{recipe_name}.yaml'
    # Fallback path for the recipe would be TURBINIA_CLI_CONFIG_PATH/recipe_name
    # This is the same path where the client configuration is loaded from.
    recipe_path_fallback = os.path.expanduser(ctx.obj.config_path)
    recipe_path_fallback = os.path.join(recipe_path_fallback, recipe_name)

    if os.path.isfile(recipe_name):
      recipe_path = recipe_name
    elif os.path.isfile(recipe_path_fallback):
      recipe_path = recipe_path_fallback
    else:
      log.error(f'Unable to load recipe {recipe_name}.')
      return

    try:
      with open(recipe_path, 'r', encoding='utf-8') as recipe_file:
        # Read the file and convert to base64 encoded bytes.
        recipe_bytes = recipe_file.read().encode('utf-8')
        recipe_data = base64.b64encode(recipe_bytes)
    except OSError as exception:
      log.error(f'Error opening recipe file {recipe_path}: {exception}')
      return
    except TypeError as exception:
      log.error(f'Error converting recipe data to Base64: {exception}')
      return
    # We found the recipe file, so we will send it to the API server
    # via the recipe_data parameter. To do so, we need to pop recipe_name
    # from the request so that we only have recipe_data.
    request['request_options'].pop('recipe_name')
    # recipe_data should be a UTF-8 encoded string.
    request['request_options']['recipe_data'] = recipe_data.decode('utf-8')

  # Send the request to the API server.
  try:
    click.echo(f'Sending request: {request}')
    api_response = api_instance.create_request(request)
    click.echo(f'Received response: {api_response}')
  except exceptions.ApiException as exception:
    log.error(
        f'Received status code {exception.status} '
        f'when calling create_request: {exception.body}')
  except (TypeError, exceptions.ApiTypeError) as exception:
    log.error(f'The request object is invalid. {exception}')



# -*- coding: utf-8 -*-
# Copyright 2022 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Turbinia API client command-line tool."""

import unittest
import mock
import pathlib

from turbinia_client.turbiniacli import TurbiniaCli
from fastapi.testclient import TestClient

from turbinia.api.api_server import app


class TestTurbiniaCli(unittest.TestCase):
  """Turbinia-client cli tool tests."""

  def setUp(self) -> None:
    """Sets up the client for the tests."""
    self.api_client = TestClient(app)
    config_path = pathlib.Path(__file__).parent.parent
    self.client = TurbiniaCli(
        config_instance='default', config_path=config_path)

  def testInitialization(self) -> None:
    """Tests if the client was initialized properly."""
    self.assertIsInstance(self.client, TurbiniaCli)

  def testReadConfiguration(self) -> None:
    """Tests the client reads the configuration file properly."""
    self.assertEqual(self.client.config_dict, {})
    self.client.read_api_configuration()
    self.assertIn('API_SERVER_ADDRESS', self.client.config_dict)
    self.assertIn('API_SERVER_PORT', self.client.config_dict)
    self.assertIn('API_AUTHENTICATION_ENABLED', self.client.config_dict)
    self.assertIn('CLIENT_SECRETS_FILENAME', self.client.config_dict)
    self.assertIn('CREDENTIALS_FILENAME', self.client.config_dict)

  @mock.patch(
      'turbinia_api_lib.api.turbinia_evidence_api.TurbiniaEvidenceApi.get_evidence_types_with_http_info'
  )
  def testGetEvidenceArguments(self, mock_response) -> None:
    """Tests the get_evidence_arguments method."""
    test_response = self.api_client.get('/api/evidence/types')
    mock_response.return_value = test_response
    api_response = self.client.get_evidence_arguments()
    self.assertEqual(test_response, api_response)

  @mock.patch(
      'turbinia_api_lib.api.turbinia_configuration_api.TurbiniaConfigurationApi.get_request_options_with_http_info'
  )
  def testGetRequestOptions(self, mock_response) -> None:
    """Tests the get_request_options method."""
    test_response = self.api_client.get('/api/config/request_options')
    mock_response.return_value = test_response
    api_response = self.client.get_request_options()
    self.assertEqual(test_response, api_response)
